//=- ArrayDFA.h - Array data flow analysis.                               --==//
//
//===----------------------------------------------------------------------===//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// Copyright (c) 2015 Advanced Micro Devices, Inc.  All rights reserved.
// Developed by: Advanced Micro Devices, Inc
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// with the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// Redistributions of source code must retain the above copyright notice, this
// list of conditions and the following disclaimers.
// Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimers in the documentation
// and/or other materials provided with the distribution.
// Neither the names of Advanced Micro Devices, Inc., nor the names of its
// contributors may be used to endorse or promote products derived from this
// Software without specific prior written permission.
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS WITH
// THE SOFTWARE.
//
//===----------------------------------------------------------------------===//
#include "llvm/ADT/MapVector.h"
#include "llvm/ADT/PostOrderIterator.h"
#include "llvm/Analysis/ISLQueryBuilder.h"
#include "llvm/Analysis/LoopAccessAnalysis.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Analysis/MemoryBuiltins.h"
#include "llvm/Analysis/PostDominators.h"
#include "llvm/Analysis/TargetLibraryInfo.h"
#include "llvm/IR/CFG.h"
#include "llvm/IR/Dominators.h"
#include "llvm/IR/ModuleSlotTracker.h"
#include "llvm/IR/TypeFinder.h"
#include "llvm/InitializePasses.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Transforms/IPO.h"
#include "llvm/Transforms/Utils/Cloning.h"
#include "llvm/Transforms/Utils/CodeExtractor.h"
#include "llvm/Transforms/Utils/LoopUtils.h"
#include <iostream>
#include <list>

using namespace llvm;
namespace llvm {

// Forward Declaration
struct DFASet;
struct AM;       // Abbreviation for Array Marker: AM
struct ArrayAcc; // Abbreviation for ArrayAccessInfo: ArrayAcc (or AC)
struct ArrBlock; // Abbreviation for ArrBlockInfo: ArrBlock (or AB)
typedef DenseMap<Value *, Value *> PHEqValMap;

enum ArrAnalyType { ARRAY_ANALYSIS, AOP_ANALYSIS };

struct MemObjAndUse {
  Value *RootMemObj;
  std::list<Value *> *SafeUsesOfRootMemObject;
  Loop *InitLoop;
  BasicBlock *InitLoopExitBB;
  Loop *InitLoopsExitLoop;
};

struct ArrayFuncNode {
  Value *ARR;
  Function *Func;
  std::list<CallInst *> *CIList;
};

struct DeadOpInfoNode {
  BasicBlock *BB;                    // BB were a virtual def could be dead
  Loop *L;                           // Loop of BB
  DFASet *DeadDefs;                  // Dead Defs in terms of scale
  std::list<ArrayAcc *> *DeadACList; // Ac's associated with DeadDefs
};

  // Associate Basic block with DFASet
typedef DenseMap<BasicBlock *, DFASet *> BBMapDFASet;

// Utils
Value *GetPtrArg(CallInst *CI);
Value *FindIntArg(CallInst *CI);

class ArrayDFA : public ModulePass {
public:
  static char ID; // Pass identification, replacement for typeid
  void InitCommonData(Module &M);

  explicit ArrayDFA() : ModulePass(ID) {
    initializeArrayDFAPass(*PassRegistry::getPassRegistry());
  }

  // Argument info
  // 1. Module
  // 2. AnalyTy (Analysis Type)  = ARRAY_ANALYSIS for Array Analysis
  //                               AOP_ANALYSIS for AOP Analysis
  // 3. MARR is the array for ADFA
  // 4. CILst is a specific set of call instructions list used during AOP
  //    analysis only, it is null during ARR analysis
  explicit ArrayDFA(Module &M, Value *MARR,
                    ArrAnalyType AnalyTy = ARRAY_ANALYSIS,
                   std::list<CallInst *> *CILst = nullptr)
      : ModulePass(ID) {
    AOPAnalysis = AnalyTy;
    initializeArrayDFAPass(*PassRegistry::getPassRegistry());
  }

  ~ArrayDFA() {
      ClearADFADataStructures();
  }

  bool runOnModule(Module &M) override;

  virtual void getAnalysisUsage(AnalysisUsage &AU) const {
    AU.addRequired<TargetLibraryInfoWrapperPass>();
    AU.addRequired<LoopInfoWrapperPass>();
    AU.addRequired<ScalarEvolutionWrapperPass>();
    AU.addRequired<DominatorTreeWrapperPass>();
    AU.addRequired<PostDominatorTreeWrapperPass>();
    AU.setPreservesAll();
  }

  static Type *LTy;
  ScalarEvolution *SE;
  LoopInfo *LI;
  const DataLayout *DL;

  // Main Array or AOP which is currently analyzed.
  Instruction *MARR;

  // Size of each element in MARR in bytes.
  unsigned int MARRElemSz;

  // SCEV expression for MARRElemSz
  const SCEV *MARRElemSzSV;

  // MainAnlyLoop could be the only region where MARR is accessed
  // in which case DFA could be restricted to MainAnlyLoop alone.
  Loop *MainAnlyLoop;

  // if MainAnlyLoop does not exist then DFA would be restricted to
  // MainAnlyFn.
  Function *MainAnlyFn;

  Value *VirtualUK; // Virtual Unknown to handle the case where SV is null.
  PHEqValMap PHIEqValMap;
  int AOPAnalysis;
  BBMapDFASet UseSetMap;
  BBMapDFASet DefSetMap;
  BBMapDFASet
      OrgUseSetMap; // Redundant data structure to track original Uses.
  BBMapDFASet
      OrgDefSetMap; // Redundant data structure to track original Defs.
  BBMapDFASet InSetMap;
  BBMapDFASet OutSetMap;

  std::list<ArrBlock *> Scale;
  int ScaleLen;
  int SetSz;
  static bool isValueLessThanEqual(AM &Fam, AM &Sam,
                                   std::list<ArrBlock *> &Scale);
  static Loop *isInsnInLoop(LoopInfo *LI, Instruction *Insn);
  static const SCEV *GetImmLoopExitValue(LoopInfo *LI,
                                         ScalarEvolution *SE, ArrayAcc *AC);
  static int GetTySize(Type *Ty, const DataLayout *DL);
  bool isIterativeAdd(ArrayAcc *AC);
  static bool isPHIEqual(PHEqValMap &PHIEqValMap, PHINode *PHI1, PHINode *PHI2);
  static int GetIncVal(LoopInfo *LI, const SCEV *SV, int &status,
                       Instruction **AddI = nullptr);

  // Following are the external interface routines.
  std::list<DeadOpInfoNode *> DeadOpInfoNodeList;

  // Compute the liveness for an AOP or an Array.
  // M is the Module
  // MARR could be ARR or AOP
  // AnlyTy is the analysis type which is
  //       a) ARRAY_ANALYSIS for Array and
  //       b) AOP_ANALYSIS for AOP
  // CILst contain function which specifically need during AOP analysis.
  ArrayDFA *ComputeLivenessInfo(Module &M, Value *MARR, ArrAnalyType AnlyTy,
                                std::list<CallInst *> *CILst);

  // Release all the resources allocated by ArrayDFA.
  void ClearADFADataStructures(void);

  // Check if a single ArrBlock is live out of ArrayAcc
  ArrBlock *isSingleActiveLiveOut(ArrayAcc *AC);

  // Finds the LiveOuts of the given Basic block BB
  // 1. BB is the concerned BasicBlock
  // 2. LiveOutSets is dynamically created array of pointers to ArrBlock,
  //    the client is supposed to free this array.
  // 3. Len is the number of entries in LiveOutSets
  void GetLiveOut(BasicBlock *BB, ArrBlock **LiveOutSets, int &Len);

  // Finds the LiveIns of the given Basic block BB
  // 1. BB is the concerned BasicBlock
  // 2. LiveInSets is dynamically created array of pointers to ArrBlock,
  //    the client is supposed to free this array.
  // 3. Len is the number of entries in LiveOutSets
  void GetLiveIn(BasicBlock *BB, ArrBlock **LiveInSets, int &Len);

  // Finds the Uses of the given Basic block BB
  // 1. BB is the concerned BasicBlock
  // 2. UseSets is dynamically created array of pointers to ArrBlock,
  //    the client is supposed to free this array.
  // 3. Len is the number of entries in LiveOutSets
  void GetUse(BasicBlock *BB, ArrBlock **UseSets, int &Len);

  // Finds the Defs of the given Basic block BB
  // 1. BB is the concerned BasicBlock
  // 2. DefSets is dynamically created array of pointers to ArrBlock,
  //    the client is supposed to free this array.
  // 3. Len is the number of entries in LiveOutSets
  void GetDef(BasicBlock *BB, ArrBlock **DefSets, int &Len);

  // Check all criteria of pruning a function.
  // This includes ISL specifics only
  bool isLegalToPrune(Function *F, Value *StartAddr, Value *TotalElements,
                      unsigned ElementSize);

  // Display the dead ops found.
  void DisplayDeadOpInfoNodeList(void);

  // Display the IN and OUT sets computed.
  void DumpFuncsInsOuts(void);

  void DumpFuncsDefsAndOuts(void); 

  // Display the USE and DEF sets computed.
  void DumpFuncsUseDefs(void);

private:
  TargetLibraryInfo *TLI;
  const LoopAccessInfo *LAI = nullptr;
  PostDominatorTree *PDT;
  DominatorTree *DT;
  Type *I8Ty;
  Type *I8PtrTy;
  LLVMContext *Cntx;
  Instruction *LifeTimeStart;
  bool M64;
  bool M32;
  Module *Modl;

  // RootMemObject is the Mem object whose addresses are stored into
  // MARR (in case of AOP analysis). RootMemObject should not be used
  // for any other purpose apart from initializing the MARR.
  Value *RootMemObject;

  // There could be multiple RootMemObjects used in exclusive paths
  // which could arise in case of function versioning. which is the
  // case with MCF.
  std::list<MemObjAndUse *> RootMemObjectList;

  // MARROtherUses are analysed and access information of each
  // and every accesses in collected in MARRAccList.
  std::list<ArrayAcc *> MARRAccList;

  typedef DenseMap<BasicBlock *, int> PresenceBBMap;

  // Reverse order of basic block starting from an exit
  // block to speed up DFA.
  std::list<BasicBlock *> RevOrdBB;
  std::list<BasicBlock *> NormalOrdBB;

  // Edges which need to be avoided for proper ADFA.
  PresenceBBMap HinderEdges;

  typedef DenseMap<ArrayAcc *, int> AccessIOrdMap;
  AccessIOrdMap AccessIOrderMap;

  typedef DenseMap<Loop *, int> PresenceLoopMap;

  // Loops in which uses and defs cannot be moved to preheaders
  PresenceLoopMap ConservativeLoops;
  PresenceBBMap AvoidPreHeadBlocks;

  typedef DenseMap<Value *, int> PresenceValMap;
  PresenceValMap VistedValMap;

  // This is a list of functions which determine the
  // the criteria for selecting an AOP for analysis.
  // This is implemented for the specific needs of MCF.
  std::list<CallInst *> *StartCIList;

  Type *GetBasePtrTyAndIdLevel(Type *Ty, int &IdLevel);
  bool isRootMemObjUsedInMAOPInitOnly(void);
  Type *GetBaseArrTy(Type *Ty);
  void DumpSafeUsesOfRootMemObject(void);
  bool isLifeTimePattern(Instruction *MainI);
  const SCEV *GetLoopIterCount(Instruction *MainGEPSVI);
  bool GetReverseOrderedBB(Function *F, Loop *L);
  void GetExitBB(Function *F, std::list<BasicBlock *> &ExitBBList);
  BasicBlock *GetExitBB(Function *F);
  bool isFuncOfReqType(Function *Fn);
  void ClearFuncAnalysisInfo(void);
  bool AllLoopAreSame(std::list<Loop *> &AllLoops);

  void AddToMARRAccList(Instruction *AccessI, Instruction *ArrAddrI);
  void PerformAOP_DFA(Module &M, Instruction *MAOP1,
                                std::list<CallInst *> *CILst);
  void DisplayMARRAccList(Instruction *MArr, int DispKey,
                          bool DispLoop = false);
  bool FindAnalysisScopeOfMARR(void);
  void MergeSVOfUnRolledAccesses(void);
  Loop *isInsnInLoopDFA(Instruction *Insn);
  bool ComputeSVEnd(ArrayAcc *AC);
  bool ComputeSVStartEndStep(void);
  bool ComputeRelationalStart(ArrayAcc *AC);
  void ComputeMARRAccSetScale(void);
  int SearchArrSetScale(ArrayAcc *AC);
  void ComputeUseDefs(void);
  BasicBlock *GetPreHeadBlock(ArrayAcc *AC);

  void ComputeMemLiveness(void);
  void RelateUseDefs(void);
  bool isSortedMergeKeysContinuous(int MergedKeys[], int MergedKeysLen);
  void EraseRedundantKnownSetEntries(void);
  AM GetAlreadyExistingEqArrayMarker(AM AM);
  void ComposeSubKeyInAllDFASetsHavingSuperKey(int SuperKey, int SubKey);
  void MergeSameSVOfUnRolledAccesses(void);
  void PerformARR_DFA(Module &M, Instruction *MArr);
  const SCEV *ReplaceUnknownWithSV(const SCEV *MainSV, const SCEVUnknown *U1,
                                   const SCEV *SV_Index);
  void ComposeSubKeyInAllDFASetsHavingSuperKeyUnknown(int SuperKey, int SubKey);
  void MergeSVOfUnRollAccesses(void);
  void
  PerformSimpleMerge(ScalarEvolution *SE, std::list<ArrayAcc *>::iterator ACCI,
                          std::list<ArrayAcc *> &DelDFAList);
  void isAdjVectorBBMerge(std::list<ArrayAcc *>::iterator ACCI,
                          std::list<ArrayAcc *> &DelDFAList);
  void isAdjVectorLoopMerge(ScalarEvolution *SE, std::list<ArrayAcc *>::iterator ACCI,
                            std::list<ArrayAcc *> &DelDFAList);
  ArrayAcc *GetAdjVectorLoopAC(BasicBlock *ExitB,
                               std::list<ArrayAcc *> &DelDFAList);
  void ReverseBackWardAccess(void);
  int GetNumSucc(BasicBlock *BB);
  int GetNumPred(BasicBlock *BB);
  void RememberHinderEdges(BasicBlock *PreHeader);
  ArrayAcc *isAccessBBAlreadyRecorded(BasicBlock *BB);
  ArrayAcc *isAccessIAlreadyRecorded(Instruction *NewI);
  void SortMARRAccList(void);
  bool IsVecMiddleBlock(BasicBlock *ExitB);
  BasicBlock *isPrimLoop(BasicBlock *BB);
  bool IsVecPhBlock(BasicBlock *ExitB);
  void RecordInfoOnDeadDefs(void);
  void CollectDeadACInfo(BasicBlock *CurrBB, DFASet *DeadDefs);
  void GetKeysInSet(DFASet *TSet, int Keys[], int &NumKeys);

  void ComputeActiveLiveOutsOfAC(void);
  bool IsASucc(BasicBlock *BB, BasicBlock *SuccOfBB);
  bool PathExists(BasicBlock *BBStart, BasicBlock *BBDest);
  MemObjAndUse *GetExistingRootMemObj(Value *NewRoot);
  void DisplayRootMemObjects(void);
  void SkipRawUsesofRootMemObjects(void);
  void AnalyzeAllMARRUses(Instruction *ValI, int &Status);
  void AnalyzeArrEleUsage(Instruction *MemAccessI);
  bool AnalyzeStructFieldUsage(Instruction *LoadAddr);

  Instruction *GetAddrValAOP(Value *Val);
  bool isCallInStartCIList(Instruction *I);
  Instruction *FindCurrentRootMemObj(Instruction *StoreI,
                                     std::list<Instruction *> &ImmediateUse);
  bool FindTrackCurrentRootMemObj(Instruction *SI);

  Instruction *GetAddrValArr(Instruction *Val);
  void TransferAccessI(ArrayAcc *DestAC, ArrayAcc *SrcAC,
                       std::list<ArrayAcc *> &DelDFAList);

  bool IPostDominatesLoopEntry(Instruction *I);
  bool isInLoopExitBlocks(Loop *CurrLoop, BasicBlock *BB);
  BasicBlock *GetLoopEntryBB(Loop *CurrLoop);
  void FindConservativeLoops(void);
  void PerformSimpleMergeBB(ScalarEvolution *SE, std::list<ArrayAcc *>::iterator ACCI,
                            std::list<ArrayAcc *> &DelDFAList);
  bool CheckForCompleteWordAcc(ScalarEvolution *SE);
  BasicBlock *DetermineVecConPreHead(BasicBlock *BB);
  BranchInst *GetBranchI(BasicBlock *iBB);
  bool GetMAOPInitLoop(MemObjAndUse *RN);
  bool isSingleNodeLoop(BasicBlock *BB);
  void ReSolvePHIs(BasicBlock *DummyPreHead, BasicBlock *PreHeader);
  void FreeADFASets(void);
  void ClearFuncsWithArrInp(void);
  void ClearMAOPAccList(void);
  void ClearRootMemObjectList(void);
  void ClearDeadOpInfoNodeList(void);
  bool trackUsageOfInterest(Value *V, Function *F,
                            SmallVector<Value *, 2> &UseInst,
                            SmallVector<CallInst *, 2> &RecursiveCS,
                            MapVector<Value *, bool> &Tracker);
  void GetActiveABs(DFASet *DSet, ArrBlock *ABS[], int &Len);
};


// ========================================================================================
// Start: DFASet is an abstract data type to simulate a Set (is used to manage
// USE/DEF/IN/OUT) Here each bit 'AccSet' indicates the presence of absence of
// an element (as indicated by 'Scale[i]'). This helps to compute
// Set Unions/Intersection/Differences faster.
// ========================================================================================

struct DFASet {

  char *AccSet; // Each bit in AccSet corresponds to corresponding element in
                // Scale

  static ArrBlock *
  GetNthArrBlockInScale(std::list<ArrBlock *> &Scale, int PosKey);
  static void DisplayScale(std::list<ArrBlock *> &Scale);

   DFASet(int SetSz) { AccSet = (char *)calloc(SetSz, sizeof(char)); }
  ~DFASet() { free(AccSet); }

  friend DFASet *DFASetUnion(DFASet *S1, DFASet *S2, int SetSz);
  friend DFASet *DFASetDifference(DFASet *S1, DFASet *S2, int SetSz);
  friend bool isDAFSetEqual(DFASet *S1, DFASet *S2, int SetSz);
  void ComposeSet(int Key);
  void UnComposeSet(int Key);
  bool isBitSet(int Key);
  bool isDFASetEmpty(int SetSz);
  void Display(std::list<ArrBlock *> &Scale, int ScaleLen, int SetSz);
  void GetAllKeysInDFASet(int Keys[], int &NumKeys, int ScaleLen);
  bool GetNumBitsSet(int ScaleLen);
};

// ========================================================================================
// End: DFASet
// ========================================================================================




// ========================================================================================
// Start: AM (Array Marker) is an abstract data type to track a location in ARR.
// ========================================================================================
struct AM { // Abbreviation: AM Stands for Array Marker

  // Contains Scev Expr of ARR loc, if Arr_Index is non constant
  const SCEV *SV_Index;

  friend int isConsAMEqual(AM const &, AM const &);
  //friend int operator<=(AM const &, AM const &);
  friend int isAMLessThanEq(AM const &, AM const &,
                            std::list<ArrBlock *> &Scale, ScalarEvolution *SE);
  friend int isAMEqual(AM &Fam, AM &Sam, ScalarEvolution *SE);

  const SCEV *GetSV_Index(void);
  void SetSV_Index(const SCEV *SV_Idx);
  bool isConst(void) const;
  bool isVar(void) const;
  bool isUK(void) const;
  long GetConVal(void) const;
  void Display(void);
};

// ========================================================================================
// End: AM (Array Marker)
// ========================================================================================

// ========================================================================================
// Start: ArrBlock is an abstract data type to track a block of locations in
// ARR.
// ========================================================================================
struct ArrBlock {
  AM StartAM; // Start of ArrBlock
  AM EndAM;   // End of ArrBlock
  AM StepAM;  // Step size for array accesses happen from StartAM to EndAM
  AM WordSz;  // Word size indicates the size of word accessed

  friend struct DFASet;

  ArrBlock(AM Start_SVI, AM End_SVI, AM StepSzI, AM WrdSz) {
    StartAM = Start_SVI;
    EndAM = End_SVI;
    StepAM = StepSzI;
    WordSz = WrdSz;
  }

  AM *GetStartAM(void);
  AM *GetEndAM(void);
  AM *GetStepAM(void);
  AM *GetWordSz(void);
  void SetStartAM(const SCEV *SV_Idx);
  void SetEndAM(const SCEV *SV_Idx);
  void SetStepAM(const SCEV *SV_Idx);
  void SetWordSz(const SCEV *WSz);
  bool isWordSzAndStepSzSame(ScalarEvolution *SE);
  bool isACompleteWord(ScalarEvolution *SE);

  friend int isABEqual(ArrBlock const &, ArrBlock const &, ScalarEvolution *SE);
  bool isStartAMConst(void);
  bool isEndAMConst(void);
  bool isStepAMConst(void);

  bool isStartAMVar(void);
  bool isEndAMVar(void);
  bool isStepAMVar(void);
  bool isStartAMUK(void);
  bool isEndAMUK(void);
  bool isStepAMUK(void);

  bool isConst(void);
  void Display(void);
  long GetStepVal(void);
  void ReComputeWordSz(ScalarEvolution *SE);

  int ScaleKey; // An internal Data, Key index into ArrayDFA::Scale
};

// ========================================================================================
// End: ArrBlock
// ========================================================================================


// ========================================================================================
// Start: ArrayAcc is an abstract data type to track array accesses. It stores
// all the required information to perform ADFA.
// ========================================================================================
struct ArrayAcc {
  Instruction *ArrayAddr; // Pointer arith instruction
  const SCEV *ArrAddrSV;  // Scev for ArrayAddr
  ArrBlock AB;            // Block of elements which are accessed.

  // Array Access Insns like Loads/Stores/Calls operating on ArrayAddr
  // Note: Until later merge activity AccessInsList would contain a
  // single instruction.
  std::list<Instruction *> AccessInsList;
  BasicBlock *PreHeadBB; // A block where AccessI is assumed to be use/def

  DFASet *ActiveOutList; // Analysed and collected information on AB's which
                         // could be live outs conservatively.

  ArrayAcc(Instruction *ArrayAddrI, Instruction *ArrayAccI, ScalarEvolution *SE)
      : AB{{0}, {0}, {0}, {0}} {

    ArrayAddr = ArrayAddrI;
    ArrAddrSV = SE->getSCEV(ArrayAddrI);
    AB = ArrBlock{{0}, {0}, {0}, {0}};
    AccessInsList.push_back(ArrayAccI);
    AB.ScaleKey = -1;
    ActiveOutList = nullptr;
    PreHeadBB = nullptr;
  }

  bool isConst(void);

  bool isInAccessInsList(Instruction *I);
  bool AccessInsListHasStore(void);
  void Display(LoopInfo *LI, int DispKey, int DispCount, bool DispLoop = false);
  Instruction *GetFirstIAccessInsList(void);
  int GetAccessIListLen(void);
  bool ComputeSVStart(ScalarEvolution *SE, Instruction *MARR);
  bool ComputeSVStepSz(LoopInfo *LI, ScalarEvolution *SE, const DataLayout *DL);
  bool ComputeSVEnd(LoopInfo *LI, ScalarEvolution *SE, Instruction *MARR,
                    Value *VirtualUK,
                    unsigned int MARRElemSz);
  bool ComputeRelationalStart(LoopInfo *LI,
                              ScalarEvolution *SE, PHEqValMap &PHIEqValMap);
  const SCEV *GetLoopIterCount(LoopInfo *LI,
                               ScalarEvolution *SE, Instruction *MainGEPSVI);
  const SCEV *GetAccWordSz(ScalarEvolution *SE, const DataLayout *DL);
  Value *GetPossibleVarStart(LoopInfo *LI, Value *temp, Instruction *LoopI);
  Value *GetValAtLoopEntry(LoopInfo *LI, PHINode *PHIN, Loop *L);
  Loop *FindCommonLoop(LoopInfo *LI, PHINode *PHIN);
  Value *AnalyzeLoopEntryExitPHI(const Value *temp);
  Value *GetEquivalentExitingConcreteVals(Value *temp);
  void IsPHIConCurrent(PHEqValMap &PHIEqValMap, Value * ValA, Value *ValB,
                       std::list<Value *> &VisitedPHIA,
                       std::list<Value *> &VisitedPHIB, int &status,
                       Value **EqVal);
  bool HaveVisitedPHI(std::list<Value *> &VisitedPHIX, Value *PHINX);
  bool isBasicAdd(Value *ValA, Value *ValB);
  bool isVectorisedLoop(unsigned int MARRElemSz);
  BasicBlock *GetFirstIParent(void);
  void FindEqVals(PHEqValMap &PHIEqValMap, Value *temp);
  void TraverseConCurrentPHIs(PHEqValMap &PHIEqValMap, Value * PHI);
  bool AllOpsOfPHISame(ScalarEvolution *SE, PHINode *PHIN);
  unsigned GetFirstIOpcode(void);
};

// ========================================================================================
// End: ArrayAcc
// ========================================================================================

} // namespace llvm
