//===- llvm/Analysis/PerfectLoopNests.h - Populate Nests --------*- C++ -*-===//
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file tries to populate perfect loop nests from LoopInfo.
// A loop nest is perfect when all the computable instructions are within the
// innermost loop except for loop controlling instructions for outer loops.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_ANALYSIS_PERFECTLOOPNESTS_H
#define LLVM_ANALYSIS_PERFECTLOOPNESTS_H

#include "llvm/Analysis/IVDescriptors.h"
#include "llvm/Analysis/LoopPass.h"
#include "llvm/Analysis/ScalarEvolution.h"
#include "llvm/ADT/SmallVector.h"
#include "llvm/IR/InstrTypes.h"
#include "llvm/IR/Instructions.h"
#include "llvm/IR/IntrinsicInst.h"

namespace llvm {
class Loop;
class LoopInfo;
class ScalarEvolution;
class SCEV;

/// Return true if the loop @L is rotated, otherwise false.
bool IsRotatedLoop(Loop *L);

/// Return an induction variable in loop @L.
PHINode *GetIndVar(const Loop *L, ScalarEvolution *SE);

/// Unrotated loops form perfect nests only if loop's header and latch contain
/// only loop control updates and exit check. Any other instruction is deemed
/// to violate perfect nesting. This will also populate the different loop
/// mappings.
bool HasSimpleHeaderLatch(Loop *L, ScalarEvolution *SE);

// Ordered list of loops from outer loop to the innermost loop that form the
// loop nest.
typedef SmallVector<Loop *, 2> LoopNestType;

/// Populate perfect loop nests.
class PerfectLoopNests {

private:
  std::vector <LoopNestType> PerfectLoopNests;

  // Return true if all blocks of @L are perfectly nested under @L's subloop.
  // @L's header and latch are exempted though.
  bool BlocksPerfectlyNestedUnder(Loop *L);

public:
  std::vector<LoopNestType> &getPerfectLoopNests() { return PerfectLoopNests; }

  // Optional hook to build perfect nests just under a loop @L.
  bool PopulatePerfectLoopNestsUnder(Loop *L, ScalarEvolution *SE);

  // Bulk builder of perfects nests under @LI.
  void PopulatePerfectLoopNestsUnder(LoopInfo *LI, ScalarEvolution *SE);

  // Clear PerfectLoopNests of its contents.
  void clear();

  // Print the contents of PerfectLoopNests.
  void print();

  // Print only a single loop nest @LN.
  static void printLoopNest(const LoopNestType &LN);
};

/// Find @L's depth in @LN.
int DepthInNest(const LoopNestType &LN, const Loop *L);

} // end namespace llvm

#endif // LLVM_ANALYSIS_PERFECTLOOPNESTS_H
