//===- LegacyPassManager.h - Legacy Container for Passes --------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
// Modifications Copyright (c) 2019 Advanced Micro Devices, Inc. All rights reserved.
// Notified per clause 4(b) of the license.
//
//===----------------------------------------------------------------------===//
//
// This file defines the legacy PassManager class.  This class is used to hold,
// maintain, and optimize execution of Passes.  The PassManager class ensures
// that analysis results are available before a pass runs, and that Pass's are
// destroyed when the PassManager is destroyed.
//
//===----------------------------------------------------------------------===//

#ifndef LLVM_IR_LEGACYPASSMANAGER_H
#define LLVM_IR_LEGACYPASSMANAGER_H

#include "llvm/Pass.h"
#include "llvm/Support/CBindingWrapping.h"
#include "llvm/Support/CommandLine.h"

namespace llvm {

class Pass;
class Module;

namespace legacy {

class PassManagerImpl;
class FunctionPassManagerImpl;

/// PassManagerBase - An abstract interface to allow code to add passes to
/// a pass manager without having to hard-code what kind of pass manager
/// it is.
class PassManagerBase {
public:
  virtual ~PassManagerBase();

  /// Add a pass to the queue of passes to run.  This passes ownership of
  /// the Pass to the PassManager.  When the PassManager is destroyed, the pass
  /// will be destroyed as well, so there is no need to delete the pass.  This
  /// may even destroy the pass right away if it is found to be redundant. This
  /// implies that all passes MUST be allocated with 'new'.
  virtual void add(Pass *P) = 0;
};

/// PassManager manages ModulePassManagers
class PassManager : public PassManagerBase {
public:

  PassManager();
  ~PassManager() override;

  void add(Pass *P) override;

  /// run - Execute all of the passes scheduled for execution.  Keep track of
  /// whether any of the passes modifies the module, and if so, return true.
  bool run(Module &M);

private:
  /// PassManagerImpl_New is the actual class. PassManager is just the
  /// wraper to publish simple pass manager interface
  PassManagerImpl *PM;
};

/// FunctionPassManager manages FunctionPasses.
class FunctionPassManager : public PassManagerBase {
public:
  /// FunctionPassManager ctor - This initializes the pass manager.  It needs,
  /// but does not take ownership of, the specified Module.
  explicit FunctionPassManager(Module *M);
  ~FunctionPassManager() override;

  void add(Pass *P) override;

  /// run - Execute all of the passes scheduled for execution.  Keep
  /// track of whether any of the passes modifies the function, and if
  /// so, return true.
  ///
  bool run(Function &F);

  /// doInitialization - Run all of the initializers for the function passes.
  ///
  bool doInitialization();

  /// doFinalization - Run all of the finalizers for the function passes.
  ///
  bool doFinalization();

private:
  FunctionPassManagerImpl *FPM;
  Module *M;
};

} // End legacy namespace

// Create wrappers for C Binding types (see CBindingWrapping.h).
DEFINE_STDCXX_CONVERSION_FUNCTIONS(legacy::PassManagerBase, LLVMPassManagerRef)

// AOCC begin
#if AOCC_DEVELOPER_BUILD
extern cl::opt<bool> VerifyAllPasses, VerifyAllPassesP, VerifyAllPassesX, VerifyAllPassesA;
extern cl::opt<std::string> VerifyAllControl;

#define VERIFY_ENV (getenv("LLVM_VERIFY_ALL"))
#define VERIFY_OPT (bool(VerifyAllPasses))
#define DO_VERIFY (VERIFY_OPT || VERIFY_ENV)
#define VERIFY_FLAG_ON(FF) ((VERIFY_OPT && VerifyAllControl.find(FF) != std::string::npos) || \
                            (!VERIFY_OPT && VERIFY_ENV && strchr(VERIFY_ENV, FF)))
#define VERIFY_FLAG_ON_OPT(FF, IRFLAG) (VERIFY_FLAG_ON(FF) || (VERIFY_OPT && llvm::IRFLAG))
#define DO_PRINT_ON_FAIL VERIFY_FLAG_ON_OPT('p', VerifyAllPassesP)
#define DO_EXIT_ON_FAIL  VERIFY_FLAG_ON_OPT('x', VerifyAllPassesX)
#define DO_VERIFY_ALWAYS VERIFY_FLAG_ON_OPT('a', VerifyAllPassesA)

#define CHECK_PASS_FS(C, F, PS) \
  if (DO_VERIFY && (DO_VERIFY_ALWAYS || (C)) && verifyFunction(F, &dbgs())) { \
    if (DO_PRINT_ON_FAIL) { \
      dbgs() << "VERIFY FAIL FOR: " << (F).getName() << ",PASS:" << PS << '\n'; \
      dbgs() << "MODULE: " << *((F).getParent()) << '\n'; \
    }; \
    if (DO_EXIT_ON_FAIL) assert(0); \
  } else { }

#define CHECK_PASS_F(C, F, P) CHECK_PASS_FS(C, F, (P)->getPassName())

#define CHECK_PASS_M(C, M, P) \
  if (DO_VERIFY && (DO_VERIFY_ALWAYS || (C)) && verifyModule(M, &dbgs())) { \
    if (DO_PRINT_ON_FAIL) { \
      dbgs() << "VERIFY FAIL FOR: " << (M).getName() << ",PASS:" << P->getPassName() << '\n'; \
      dbgs() << "MODULE: " << (M) << '\n'; \
    }; \
    if (DO_EXIT_ON_FAIL) assert(0); \
  } else { }

#define CHECK_M(C, M) \
  if (DO_VERIFY && (DO_VERIFY_ALWAYS || (C)) && verifyModule(M, &dbgs())) { \
    if (DO_PRINT_ON_FAIL) { \
      dbgs() << "VERIFY FAIL FOR: " << (M).getName() << '\n'; \
      dbgs() << "MODULE: " << (M) << '\n'; \
    }; \
    if (DO_EXIT_ON_FAIL) assert(0); \
  } else { }
#else
#define CHECK_M(C,M)
#define CHECK_PASS_M(C,M,P)
#define CHECK_PASS_F(C,F,P)
#define CHECK_PASS_FS(C,F,S)
#endif
// AOCC end
} // End llvm namespace

#endif
